//
//  CalendarViewModel.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 11/4/21.
//

import UIKit

class CalendarViewModel {
  enum Mode: Int {
    case week = 0
    case month = 1
  }

  private let repository = CalendarRepository.shared
  private let shoppingListGenerator = AutomaticListGenerator.shared

  private(set) var mode = Mode(rawValue: UserDefaults.calendarMode) ?? .month {
    didSet {
      modeDidUpdate?(mode)
    }
  }

  private(set) var calendarData: [CalendarSection] = []
  private(set) var planningData: [Section] = [] {
    didSet {
      planningDidUpdate?()
    }
  }

  var planningDidUpdate: (() -> Void)?

  var modeDidUpdate: ((Mode) -> Void)?

  private(set) var currentDay: CalendarDay?

  deinit {
    NotificationCenter.default.removeObserver(
      self,
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  init() {
    NotificationCenter.default.addObserver(
      self,
      selector: #selector(modelDidChange),
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  func fetchPlanningData(in date: Date) {
    repository.fetch(in: date) { [weak self] day in
      guard let self = self else { return }

      if day != nil {
        self.currentDay = day
      } else {
        self.currentDay = self.repository.createNewDay(for: date)
      }

      self.postPlanningData()
    }
  }

  private func postPlanningData() {
    guard currentDay != nil else { return }

    // avoid removed articles and dishes
    // swiftlint:disable:next force_unwrapping
    let availableItems = currentDay!.itemsArray.filter { $0.dish != nil || $0.article != nil }

    let lunchItems = availableItems.filter { $0.type == CalendarType.lunch.rawValue }

    let lunchSection = Section(
      title: NSLocalizedString("Lunch", comment: ""),
      items: lunchItems)

    let dinnerItems = availableItems.filter { $0.type == CalendarType.dinner.rawValue }

    let dinnerSection = Section(
      title: NSLocalizedString("Dinner", comment: ""),
      items: dinnerItems)

    let othersItems = availableItems.filter { $0.type == CalendarType.other.rawValue }

    let othersSection = Section(
      title: NSLocalizedString("Others", comment: ""),
      items: othersItems)

    planningData = [lunchSection, dinnerSection, othersSection]
  }

  func fetchCalendarData() {
    switch mode {
    case .week:
      return fetchWeeklyData()
    case .month:
      return fetchMonthlyData()
    }
  }

  func fetchWeeklyData() {
    let date = Date()
    let days = date.getDaysOfWeek()
    var weekDays: [CalendarRepresentable] = []

    days.forEach { day in
      let newDay = CalendarRepresentable(
        date: day,
        isActive: Calendar.current.isDateInToday(day))
      weekDays.append(newDay)
    }

    calendarData = [CalendarSection(items: weekDays)]
  }

  func fetchMonthlyData() {
    let date = Date()
    var sections: [CalendarSection] = []
    var weekDays: [CalendarRepresentable] = []

    var daysOfMonth = date.getDaysOfWeek() // for day names
    daysOfMonth.append(contentsOf: date.getDaysOfMonth())

    daysOfMonth.forEach { newDate in
      let newDay = CalendarRepresentable(
        date: newDate,
        isActive: Calendar.current.isDateInToday(newDate),
        isDisabled: date.monthNumber != newDate.monthNumber)

      weekDays.append(newDay)
    }

    let chunkedInWeeks = weekDays.chunked(into: 7)

    chunkedInWeeks.forEach { weekItems in
      sections.append(CalendarSection(items: weekItems))
    }

    calendarData = sections
  }

  func setMode(to mode: Mode) {
    UserDefaults.calendarMode = mode.rawValue
    self.mode = mode
  }

  // MARK: - Calendar actions
  func didSelectCalendarItem(at indexPath: IndexPath) {
    switch mode {
    case .week:
      selectWeekDay(at: indexPath)
    case .month:
      selectMonthDay(at: indexPath)
    }
  }

  private func selectWeekDay(at indexPath: IndexPath) {
    guard calendarData.count == 1 else { return }
    guard indexPath.item < calendarData[0].items.count else { return }

    calendarData[0].items.indices.forEach { calendarData[0].items[$0].isActive = false }

    calendarData[0].items[indexPath.item].isActive = true
    fetchPlanningData(in: calendarData[0].items[indexPath.item].date)
  }

  private func selectMonthDay(at indexPath: IndexPath) {
    guard indexPath.section > 0 else { return } // first row: week day names
    guard indexPath.section < calendarData.count else { return }
    guard indexPath.item < calendarData[indexPath.section].items.count else { return }

    calendarData.indices.forEach { sectionIndex in
      calendarData[sectionIndex].items.indices.forEach { itemIndex in
        calendarData[sectionIndex].items[itemIndex].isActive = false
      }
    }

    calendarData[indexPath.section].items[indexPath.item].isActive = true
    fetchPlanningData(in: calendarData[indexPath.section].items[indexPath.item].date)
  }

  @objc private func modelDidChange() {
    if let day = currentDay {
      // swiftlint:disable:next force_unwrapping
      fetchPlanningData(in: day.date!)
    }
  }

  func delete(item: CalendarItem) {
    // swiftlint:disable force_unwrapping
    repository.delete(element: item, in: currentDay!)
    shoppingListGenerator.didChangeContent(inDate: currentDay!.date!)
    // swiftlint:enable force_unwrapping
  }
}
