//
//  NewItemCalendarViewModel.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 1/5/21.
//

import UIKit

class NewItemCalendarViewModel {
  enum AddType {
    case lists
    case dishes
    case calendar
  }

  enum SelectionType {
    case article
    case dish
  }

  enum SortBy {
    case alphabetically
    case color
  }

  private let articlesRepository = ArticlesRepository.shared
  private let dishesRepository = DishesRepository.shared
  private let calendarRepository = CalendarRepository.shared
  private let shoppingListRepository = ShoppingListRepository.shared
  private(set) var addType: AddType = .calendar

  private var filter = "" {
    didSet {
      contentDidUpdate?()
    }
  }

  private(set) var selectionType: SelectionType = .article {
    didSet {
      contentDidUpdate?()
    }
  }

  private(set) var sortBy: SortBy = .alphabetically {
    didSet {
      contentDidUpdate?()
    }
  }

  /* to store selected ids. This property is only used here,
    so I don't create a property in articles or dishes entities */
  private var selectedIds: Set<UUID> = []

  var filteredItems: [AnyHashable] {
    switch selectionType {
    case .article:
      let filteredArticles: [Article]
      if filter.isEmpty {
        filteredArticles = articles
      } else {
        filteredArticles = articles.filter {
          // swiftlint:disable:next force_unwrapping
          $0.name?.lowercased() == filter.lowercased() || $0.name!.lowercased().contains(filter.lowercased())
        }
      }

      switch sortBy {
      case .alphabetically:
        // swiftlint:disable:next force_unwrapping
        return filteredArticles.sorted { $0.name!.lowercased() < $1.name!.lowercased() }
      case .color:
        return filteredArticles.sorted { $0.wrappedColor.hashValue < $1.wrappedColor.hashValue }
      }
    case .dish:
      var filteredDishes = dishes.filter { $0.name != nil }
      if !filter.isEmpty {
        filteredDishes = filteredDishes.filter {
          // swiftlint:disable:next force_unwrapping
          $0.name?.lowercased() == filter.lowercased() || $0.name!.lowercased().contains(filter.lowercased())
        }
      }

      switch sortBy {
      case .alphabetically:
        // swiftlint:disable:next force_unwrapping
        return filteredDishes.sorted { $0.name!.lowercased() < $1.name!.lowercased() }
      case .color:
        return filteredDishes.sorted { $0.wrappedColor.hashValue < $1.wrappedColor.hashValue }
      }
    }
  }

  var title: String {
    switch addType {
    case .lists, .dishes:
      return NSLocalizedString("Articles", comment: "")
    case .calendar:
      return calendarItem?.formattedDate ?? ""
    }
  }

  private(set) var currentDish: Dish?
  private(set) var calendarItem: CalendarDay?
  private var calendarType: CalendarType = .none
  private (set) var shoppingList: ShoppingList?

  private var articles: [Article] = [] {
    didSet {
      contentDidUpdate?()
    }
  }

  private var dishes: [Dish] = [] {
    didSet {
      contentDidUpdate?()
    }
  }

  var contentDidUpdate: (() -> Void)?

  deinit {
    NotificationCenter.default.removeObserver(
      self,
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  init(withDish dish: Dish?) {
    addType = .dishes
    fetchArticles()
    currentDish = dish
  }

  init(withDay calendarDay: CalendarDay, andType type: CalendarType) {
    addType = .calendar
    fetchArticles()
    fetchDishes()
    calendarItem = calendarDay
    calendarType = type
    commonInit()
  }

  init(withList: ShoppingList) {
    addType = .lists
    fetchArticles()
    shoppingList = withList
    commonInit()
  }

  private func commonInit() {
    NotificationCenter.default.addObserver(
      self,
      selector: #selector(modelDidChange),
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  private func fetchArticles() {
    articlesRepository.fetch { [weak self] articles in
      self?.articles = articles
    }
  }

  private func fetchDishes() {
    dishesRepository.fetch { [weak self] dishes in
      self?.dishes = dishes
    }
  }

  func setSelectionType(to type: SelectionType) {
    selectionType = type
  }

  func setSort(by type: SortBy) {
    sortBy = type
  }

  func filter(withText text: String) {
    self.filter = text
  }

  func didSelect(id: UUID) {
    if selectedIds.contains(id) {
      selectedIds.remove(id)
    } else {
      selectedIds.insert(id)
    }

    contentDidUpdate?()
  }

  func isSelected(id: UUID) -> Bool {
    selectedIds.contains(id)
  }

  // swiftlint:disable:next cyclomatic_complexity
  func save() {
    switch addType {
    case .dishes:
      if let dish = currentDish {
        // se añaden los artículos y se guarda
        for id in selectedIds where !dish.articlesArray.contains(where: { $0.id == id }) {
          if let article = articles.first(where: { $0.id == id }) {
            dish.addToArticles(article)
          }
        }
      }
    case .calendar:
      if let day = calendarItem {
        let dayArticles = day.itemsArray.filter { $0.article != nil && $0.type == calendarType.rawValue }
        for id in selectedIds where !dayArticles.contains(where: { $0.id == id }) {
          if let article = articles.first(where: { $0.id == id }) {
            let newItem = calendarRepository.createNewCalendarItem()
            newItem.article = article
            newItem.type = calendarType.rawValue
            day.addToItems(newItem)
          }
        }

        let dayDishes = day.itemsArray.filter { $0.dish != nil && $0.type == calendarType.rawValue }
        for id in selectedIds where !dayDishes.contains(where: { $0.id == id }) {
          if let dish = dishes.first(where: { $0.id == id }) {
            let newItem = calendarRepository.createNewCalendarItem()
            newItem.dish = dish
            newItem.type = calendarType.rawValue
            day.addToItems(newItem)
          }
        }
        calendarRepository.save()
        // swiftlint:disable:next force_unwrapping
        AutomaticListGenerator.shared.didChangeContent(inDate: day.date!)
      }

    case .lists:
      if let list = shoppingList {
        // se añaden los artículos y se guarda
        for id in selectedIds {
          if let art = list.itemsArray.first(where: { $0.article?.id == id }) {
            art.didDiscardItem = false
          } else if let article = articles.first(where: { $0.id == id }) {
            let newItem = shoppingListRepository.createItem()
            newItem.article = article
            list.addToItems(newItem)
          }
        }
      }
      return
    }
  }

  @objc private func modelDidChange() {
    fetchArticles()

    if addType == .calendar {
      fetchDishes()
    }
  }
}
