//
//  DishesViewModel.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 7/4/21.
//

import UIKit

class DishesViewModel: SimpleListViewModelProtocol {
  private var repository = DishesRepository.shared

  var groupHeight: CGFloat = 150
  var navigationItemTitle: String = NSLocalizedString("Dishes", comment: "")
  var type: ListType = .dishes

  var items: [AnyHashable] = [] {
    didSet {
      itemsDidChange?(items)
    }
  }

  var itemsDidChange: (([AnyHashable]) -> Void)?

  deinit {
    NotificationCenter.default.removeObserver(
      self,
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  init(withRepository: DishesRepository) {
    repository = withRepository
    commonInit()
  }

  init() {
    commonInit()
  }

  private func commonInit() {
    NotificationCenter.default.addObserver(
      self,
      selector: #selector(modelDidChange),
      name: Notification.Name.NSManagedObjectContextObjectsDidChange,
      object: nil)
  }

  func fetch() {
    repository.fetch { dishes in
      // swiftlint:disable:next force_unwrapping
      self.items = dishes.sorted { $0.name! < $1.name! }
    }
  }

  func delete<T>(_ item: T) {
    if let dish = item as? Dish {
      repository.delete(dish: dish)
    }
  }

  @objc private func modelDidChange() {
    repository.fetch { dishes in
      // swiftlint:disable force_unwrapping
      let validDishes = dishes.filter { $0.name != nil && !$0.name!.isEmpty }
      let sortedDishes = validDishes.sorted { $0.name!.lowercased() < $1.name!.lowercased() }
      self.items = sortedDishes
      // swiftlint:enable force_unwrapping
    }
  }

  func getSections() -> [Section] {
    guard var dishItems = items as? [Dish] else { return [] }

    let todayDish = dishItems.first { dish in
      dish.calendarItemsArray.contains { item in
        if let date = item.day?.date {
          return Calendar.current.isDateInToday(date)
        } else {
          return false
        }
      }
    }

    if todayDish != nil {
      // swiftlint:disable:next force_unwrapping
      dishItems.removeAll { $0.id == todayDish!.id }

      let firstSection = Section(title: NSLocalizedString("Today", comment: ""), items: [todayDish])
      let secondSection = Section(title: NSLocalizedString("Your dishes", comment: ""), items: dishItems)

      return [firstSection, secondSection]
    } else {
      return [Section(items: items)]
    }
  }
}
