//
//  NewArticleViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 23/4/21.
//

import UIKit

class NewArticleViewController: UIViewController {
  private lazy var nameTextView: UITextView = {
    let textView = UITextView()
    textView.layer.cornerRadius = 15
    textView.layer.masksToBounds = true
    textView.translatesAutoresizingMaskIntoConstraints = false
    textView.textColor = .systemGray
    textView.backgroundColor = .systemGray5
    textView.textContainerInset = UIEdgeInsets(top: 12, left: 10, bottom: 12, right: 10)
    textView.font = .newYorkFont(weight: .regular, andTextStyle: .body)
    textView.textContainer.maximumNumberOfLines = 2
    textView.delegate = self
    textView.text = NSLocalizedString("Give a name to your article", comment: "")
    return textView
  }()

  private lazy var deleteButton: UIButton = {
    let btn = UIButton()
    btn.setTitle(NSLocalizedString("Delete", comment: ""), for: .normal)
    btn.backgroundColor = .systemRed
    btn.layer.masksToBounds = true
    btn.layer.cornerRadius = 15
    btn.translatesAutoresizingMaskIntoConstraints = false
    btn.addTarget(self, action: #selector(deleteArticle), for: .touchUpInside)
    return btn
  }()

  private let colorsItemSelector = ColorsItemSelector()

  private var viewModel: NewArticleViewModel

  init(withViewModel: NewArticleViewModel) {
    viewModel = withViewModel
    super.init(nibName: nil, bundle: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("not happening")
  }

  override func viewDidLoad() {
    super.viewDidLoad()
    view.backgroundColor = .systemBackground
    let controllerTitle = viewModel.currentArticle != nil ? "Article" : "New article"
    navigationItem.title = NSLocalizedString(controllerTitle, comment: "")

    let rightActionTitle = viewModel.currentArticle != nil ? "Update" : "Add"

    navigationItem.rightBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString(rightActionTitle, comment: ""),
      style: .plain,
      target: self,
      action: #selector(addTapped))

    navigationItem.rightBarButtonItem?.tintColor = .label

    navigationItem.leftBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString("Cancel", comment: ""),
      style: .plain,
      target: self,
      action: #selector(cancelTapped))

    navigationItem.leftBarButtonItem?.tintColor = .label

    let tap = UITapGestureRecognizer(target: self, action: #selector(hideKeyboard))
    tap.cancelsTouchesInView = false
    view.addGestureRecognizer(tap)

    setupViews()
    setupContent()
  }

  private func setupViews() {
    let nameLabel = createNewSectionTitleLabel()
    nameLabel.text = NSLocalizedString("Name", comment: "")

    view.addSubview(nameLabel)
    view.addSubview(nameTextView)
    view.addSubview(colorsItemSelector)

    colorsItemSelector.translatesAutoresizingMaskIntoConstraints = false

    NSLayoutConstraint.activate([
      nameLabel.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor, constant: 16),
      nameLabel.leadingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leadingAnchor, constant: 16),
      nameLabel.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -16),

      nameTextView.topAnchor.constraint(equalTo: nameLabel.bottomAnchor, constant: 6),
      nameTextView.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      nameTextView.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
      nameTextView.heightAnchor.constraint(equalToConstant: 44),

      colorsItemSelector.topAnchor.constraint(equalTo: nameTextView.bottomAnchor, constant: 16),
      colorsItemSelector.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      colorsItemSelector.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
      colorsItemSelector.heightAnchor.constraint(equalToConstant: 75)
    ])

    if viewModel.currentArticle != nil {
      view.addSubview(deleteButton)
      deleteButton.topAnchor.constraint(equalTo: colorsItemSelector.bottomAnchor, constant: 44).isActive = true
      deleteButton.leadingAnchor.constraint(equalTo: view.leadingAnchor, constant: 24).isActive = true
      deleteButton.trailingAnchor.constraint(equalTo: view.trailingAnchor, constant: -24).isActive = true
      deleteButton.heightAnchor.constraint(equalToConstant: 40).isActive = true
    }
  }

  private func setupContent() {
    if let article = viewModel.currentArticle {
      nameTextView.textColor = .label
      nameTextView.text = article.name
      colorsItemSelector.setSelected(color: article.wrappedColor)
    }
  }

  private func createNewSectionTitleLabel() -> UILabel {
    let lbl = UILabel()
    lbl.font = .newYorkFont(weight: .regular, andTextStyle: .headline)
    lbl.textColor = .systemGray
    lbl.translatesAutoresizingMaskIntoConstraints = false
    return lbl
  }

  @objc private func hideKeyboard() {
    view.endEditing(true)
  }

  @objc private func addTapped() {
    viewModel.saveNewArticle(
      withName: nameTextView.text,
      andColor: colorsItemSelector.selectedColor) { [weak self] error in
      switch error {
      case .none:
        self?.dismiss(animated: true)
      case .name:
        self?.showAlertError(withMessage: NSLocalizedString("There's already an article with this name", comment: ""))
      case .noname:
        self?.showAlertError(withMessage: NSLocalizedString("The article must have a name", comment: ""))
      case .coreData:
        self?.showAlertError(withMessage: NSLocalizedString("Error storing the item", comment: ""))
      }
    }
  }

  @objc private func cancelTapped() {
    dismiss(animated: true)
  }

  @objc private func deleteArticle() {
    showConfirmationAlert(withMessage: "Do you want to delete this article?") { [weak self] in
      guard let self = self else { return }
      self.viewModel.deleteArticle()
      self.dismiss(animated: true)
    }
  }
}

extension NewArticleViewController: UITextViewDelegate {
  func textViewDidBeginEditing(_ textView: UITextView) {
    if textView.textColor != .label {
      textView.text = ""
      textView.textColor = .label
    }
  }
}
