//
//  CalendarViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 3/4/21.
//

import UIKit

private typealias CalendarDataSource = UICollectionViewDiffableDataSource<CalendarSection, CalendarRepresentable>
private typealias CalendarDataSnapshot = NSDiffableDataSourceSnapshot<CalendarSection, CalendarRepresentable>

class CalendarViewController: UIViewController {
  // swiftlint:disable implicitly_unwrapped_optional
  private let viewModel = CalendarViewModel()
  private var collectionViewHeightConstraint: NSLayoutConstraint!

  private lazy var calendarCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createLinearCalendarLayout())
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(CalendarWeekDayCell.self, forCellWithReuseIdentifier: CalendarWeekDayCell.reuseIdentifier)
    cvw.register(CalendarMonthDayCell.self, forCellWithReuseIdentifier: CalendarMonthDayCell.reuseIdentifier)
    cvw.isScrollEnabled = false
    cvw.delegate = self
    return cvw
  }()

  private lazy var planningCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createPlanningLayout())
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ArticleCell.self, forCellWithReuseIdentifier: ArticleCell.reuseIdentifier)
    cvw.register(
      TitleAndAddHeaderView.self,
      forSupplementaryViewOfKind: UICollectionView.elementKindSectionHeader,
      withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier)
    cvw.delegate = self
    return cvw
  }()

  private var calendarDataSource: CalendarDataSource!
  private var planningDataSource: ArticlesDataSource!
  // swiftlint:enable: implicitly_unwrapped_optional

  override func viewDidLoad() {
    super.viewDidLoad()
    view.backgroundColor = .systemBackground

    view.addSubview(calendarCollectionView)
    view.addSubview(planningCollectionView)

    collectionViewHeightConstraint = calendarCollectionView.heightAnchor.constraint(equalToConstant: 85)
    collectionViewHeightConstraint.isActive = true
    collectionViewHeightConstraint.priority = .defaultHigh

    NSLayoutConstraint.activate([
      calendarCollectionView.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor, constant: 8),
      calendarCollectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      calendarCollectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),

      planningCollectionView.topAnchor.constraint(equalTo: calendarCollectionView.bottomAnchor, constant: 8),
      planningCollectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      planningCollectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
      planningCollectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])

    createCalendarDataSource()
    createPlanningDataSource()

    bindViewModel()
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    tabBarController?.navigationItem.title = NSLocalizedString("Calendar", comment: "")
    tabBarController?.navigationController?.navigationBar.prefersLargeTitles = false
    setModeIcon()
  }

  private func bindViewModel() {
    viewModel.modeDidUpdate = { [weak self] mode in
      guard let self = self else { return }
      switch mode {
      case .week:
        self.calendarCollectionView.collectionViewLayout = self.createLinearCalendarLayout()
      case .month:
        self.calendarCollectionView.collectionViewLayout = self.createMonthlyCalendarLayout()
      }

      self.viewModel.fetchCalendarData()
      self.applyCalendarDataSource()
      self.setModeIcon()
    }

    viewModel.planningDidUpdate = { [weak self] in
      self?.applyPlanningDataSource()
    }

    viewModel.setMode(to: viewModel.mode)
    viewModel.fetchPlanningData(in: Date())
  }

  private func setModeIcon() {
    guard let tabController = tabBarController as? LandingViewController else { return }
    switch viewModel.mode {
    case .week:
      tabController.rightButtonImageView?.image =
        UIImage(imageLiteralResourceName: "icCalendarMonth").withRenderingMode(.alwaysTemplate)
    case .month:
      tabController.rightButtonImageView?.image =
        UIImage(imageLiteralResourceName: "icCalendarWeek").withRenderingMode(.alwaysTemplate)
    }

    tabController.rightButtonImageView?.gestureRecognizers?.removeAll()
    tabController.rightButtonImageView?.addGestureRecognizer(
      UITapGestureRecognizer(target: self, action: #selector(changeMode)))
  }

  @objc private func changeMode() {
    switch viewModel.mode {
    case .week:
      viewModel.setMode(to: .month)
    case .month:
      viewModel.setMode(to: .week)
    }
  }

  private func createCalendarDataSource() {
    // swiftlint:disable force_cast
    calendarDataSource = CalendarDataSource(
      collectionView: calendarCollectionView
    ) { [weak self] collectionView, indexPath, calendarItem -> UICollectionViewCell? in
      guard let self = self else { return nil }

      switch self.viewModel.mode {
      case .week:
        let cell = collectionView.dequeueReusableCell(
          withReuseIdentifier: CalendarWeekDayCell.reuseIdentifier,
          for: indexPath) as! CalendarWeekDayCell
        cell.config(forItem: calendarItem)
        return cell
      case .month:
        let cell = collectionView.dequeueReusableCell(
          withReuseIdentifier: CalendarMonthDayCell.reuseIdentifier,
          for: indexPath) as! CalendarMonthDayCell
        cell.config(forItem: calendarItem, isTitle: indexPath.section == 0)
        return cell
      }
    }
    // swiftlint:enable force_cast
  }

  private func applyCalendarDataSource() {
    var snapshot = CalendarDataSnapshot()
    let sections = viewModel.calendarData

    snapshot.appendSections(sections)

    for section in sections {
      snapshot.appendItems(section.items, toSection: section)
    }

    if viewModel.mode == .week {
      UIView.animate(withDuration: 0.2, delay: 0, options: .curveEaseOut) {
        self.collectionViewHeightConstraint.constant = 85
        self.view.layoutIfNeeded()
      }
    }

    calendarDataSource.apply(snapshot, animatingDifferences: true) { [weak self] in
      guard let self = self else { return }
      self.calendarCollectionView.invalidateIntrinsicContentSize()

      UIView.animate(withDuration: 0.3, delay: 0, options: .curveEaseOut) {
        self.collectionViewHeightConstraint.constant = self.calendarCollectionView.contentSize.height
        self.view.layoutIfNeeded()
      }

      if self.viewModel.mode == .week {
        let index = sections[0].items.firstIndex { $0.isActive }
        self.calendarCollectionView.scrollToItem(
          at: IndexPath(item: index ?? 0, section: 0),
          at: .centeredHorizontally,
          animated: true)
      }
    }
  }

  // swiftlint:disable:next cyclomatic_complexity
  private func createPlanningDataSource() {
    // swiftlint:disable force_cast
    planningDataSource = ArticlesDataSource(
      collectionView: planningCollectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ArticleCell.reuseIdentifier,
        for: indexPath) as! ArticleCell

      guard let calendarItem = item as? CalendarItem else { return cell }

      if let dish = calendarItem.dish {
        cell.config(with: dish)
      }

      if let article = calendarItem.article {
        cell.config(with: article)
      }

      return cell
    }

    planningDataSource.supplementaryViewProvider = { [weak self] collectionView, kind, indexPath in
      switch kind {
      case UICollectionView.elementKindSectionHeader:
        let header = collectionView.dequeueReusableSupplementaryView(
          ofKind: kind,
          withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier,
          for: indexPath) as! TitleAndAddHeaderView

        guard let self = self else { return nil }

        if let section = self.planningDataSource?.snapshot().sectionIdentifiers[indexPath.section] {
          header.configure(with: section.title)

          guard self.viewModel.currentDay != nil else { return header }
          // swiftlint:disable force_unwrapping
          switch indexPath.section {
          case 0: // lunch
            header.addAction = {
              let viewModel = NewItemCalendarViewModel(withDay: self.viewModel.currentDay!, andType: .lunch)
              let vcr = NewItemCalendarViewController(withViewModel: viewModel)
              self.present(UINavigationController(rootViewController: vcr), animated: true)
            }
          case 1: // dinner
            header.addAction = {
              let viewModel = NewItemCalendarViewModel(withDay: self.viewModel.currentDay!, andType: .dinner)
              let vcr = NewItemCalendarViewController(withViewModel: viewModel)
              self.present(UINavigationController(rootViewController: vcr), animated: true)
            }
          case 2: // others
            header.addAction = {
              let viewModel = NewItemCalendarViewModel(withDay: self.viewModel.currentDay!, andType: .other)
              let vcr = NewItemCalendarViewController(withViewModel: viewModel)
              self.present(UINavigationController(rootViewController: vcr), animated: true)
            }
          default: break
          }
          // swiftlint:enable force_unwrapping
          return header
        }

      default:
        fatalError("Implement new kind")
      }

      return nil
    }
    // swiftlint:enable force:cast
  }

  private func applyPlanningDataSource() {
    var snapshot = ArticlesDataSnapshot()
    let sections = viewModel.planningData

    snapshot.appendSections(sections)

    for section in sections {
      snapshot.appendItems(section.items, toSection: section)
    }

    planningDataSource.apply(snapshot)
  }

  private func createLinearCalendarLayout() -> UICollectionViewLayout {
    let itemSize = NSCollectionLayoutSize(
      widthDimension: .fractionalWidth(1.0),
      heightDimension: .fractionalHeight(1.0))
    let layoutItem = NSCollectionLayoutItem(layoutSize: itemSize)

    let layoutGroupSize = NSCollectionLayoutSize(
      widthDimension: .fractionalHeight(0.8),
      heightDimension: .fractionalHeight(1.0))

    let layoutGroup = NSCollectionLayoutGroup.horizontal(layoutSize: layoutGroupSize, subitems: [layoutItem])

    let section = NSCollectionLayoutSection(group: layoutGroup)
    section.orthogonalScrollingBehavior = .continuous
    section.interGroupSpacing = 12
    section.contentInsets = NSDirectionalEdgeInsets(top: 6, leading: 12, bottom: 6, trailing: 12)
    return UICollectionViewCompositionalLayout(section: section)
  }

  private func createMonthlyCalendarLayout() -> UICollectionViewLayout {
    let itemSize = NSCollectionLayoutSize(
      widthDimension: .fractionalWidth(1.0),
      heightDimension: .fractionalHeight(1.0))

    let layoutItem = NSCollectionLayoutItem(layoutSize: itemSize)
    // layoutItem.contentInsets = NSDirectionalEdgeInsets(top: 2, leading: 2, bottom: 2, trailing: 2)

    let layoutGroupSize = NSCollectionLayoutSize(
      widthDimension: .fractionalWidth(1.0),
      heightDimension: .fractionalWidth(1 / 9))

    // let layoutGroup = NSCollectionLayoutGroup.horizontal(layoutSize: layoutGroupSize, subitems: [layoutItem])
    let layoutGroup = NSCollectionLayoutGroup.horizontal(layoutSize: layoutGroupSize, subitem: layoutItem, count: 7)

    let section = NSCollectionLayoutSection(group: layoutGroup)
    return UICollectionViewCompositionalLayout(section: section)
  }

  private func createPlanningLayout() -> UICollectionViewLayout {
    return UICollectionViewCompositionalLayout.listWithHeaderLayout(groupEstimatedHeight: 70)
  }
}

extension CalendarViewController: UICollectionViewDelegate {
  func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
    if collectionView == calendarCollectionView {
      viewModel.didSelectCalendarItem(at: indexPath)
      applyCalendarDataSource()
    } else { // article or dish selected
      if let item = self.planningDataSource.itemIdentifier(for: indexPath) as? CalendarItem {
        if let article = item.article {
          let articleViewModel = NewArticleViewModel(withArticle: article)
          let newController = NewArticleViewController(withViewModel: articleViewModel)
          let navController = UINavigationController(rootViewController: newController)
          present(navController, animated: true)
        } else if let dish = item.dish {
          let dishViewModel = NewDishViewModel(withDish: dish)
          let newController = NewDishViewController(withViewModel: dishViewModel)
          let navController = UINavigationController(rootViewController: newController)
          present(navController, animated: true)
        }
      }
    }
  }

  func collectionView(
    _ collectionView: UICollectionView,
    contextMenuConfigurationForItemAt indexPath: IndexPath,
    point: CGPoint
  ) -> UIContextMenuConfiguration? {
    guard collectionView != calendarCollectionView else { return nil }

    return UIContextMenuConfiguration(
      identifier: nil,
      previewProvider: nil) {  _ in
      let deleteAction = self.deleteItemAction(at: indexPath)
      let children = [deleteAction]

      return UIMenu(title: "", image: nil, children: children)
    }
  }

  private func deleteItemAction(at indexPath: IndexPath) -> UIAction {
    let image = UIImage(systemName: "trash")

    return UIAction(
      title: NSLocalizedString("Delete", comment: ""),
      image: image,
      attributes: .destructive) { [weak self] _ in
      self?.showConfirmationAlert(withMessage: "Do you want to delete this article?") { [weak self] in
        guard let self = self else { return }
        if let item = self.planningDataSource.itemIdentifier(for: indexPath) as? CalendarItem {
          self.viewModel.delete(item: item)
        }
      }
    }
  }
}
