//
//  NewItemCalendarViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 24/4/21.
//

import UIKit

class NewItemCalendarViewController: UIViewController {
  private let searchBar = UISearchBar()
  private let topView = UIView()
  private var segmentedControl = UISegmentedControl(
    items: [NSLocalizedString("Articles", comment: ""), NSLocalizedString("Dishes", comment: "")])

  private lazy var contentCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createLayout())
    cvw.delegate = self
    cvw.contentInset.top = 10
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ArticleCell.self, forCellWithReuseIdentifier: ArticleCell.reuseIdentifier)
    return cvw
  }()

  // swiftlint:disable:next implicitly_unwrapped_optional
  private var contentDataSource: ArticlesDataSource!
  private var viewModel: NewItemCalendarViewModel

  init(withViewModel: NewItemCalendarViewModel) {
    viewModel = withViewModel
    super.init(nibName: nil, bundle: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("init(coder:) has not been implemented")
  }

  override func viewDidLoad() {
    super.viewDidLoad()

    let tap = UITapGestureRecognizer(target: self, action: #selector(hideKeyboard))
    tap.cancelsTouchesInView = false
    view.addGestureRecognizer(tap)

    view.backgroundColor = .systemBackground
    navigationItem.title = viewModel.title
    navigationController?.navigationBar.tintColor = .label

    navigationItem.rightBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString("Add", comment: ""),
      style: .plain,
      target: self,
      action: #selector(addTapped))

    if viewModel.addType == .calendar {
      navigationItem.leftBarButtonItem = UIBarButtonItem(
        title: NSLocalizedString("Cancel", comment: ""),
        style: .plain,
        target: self,
        action: #selector(cancelTapped))
    }

    viewModel.contentDidUpdate = { [weak self] in
      self?.applyDataSource()
    }

    setupSearchViews()
    setupContentViews()
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    navigationItem.leftBarButtonItem?.tintColor = .label
    navigationItem.rightBarButtonItem?.tintColor = .label
    let appearance = UINavigationBarAppearance()
    appearance.backgroundColor = .systemBackground
    appearance.backgroundImage = UIImage()
    appearance.shadowImage = UIImage()
    UINavigationBar.appearance().scrollEdgeAppearance = appearance
    UINavigationBar.appearance().standardAppearance = appearance
    UINavigationBar.appearance().compactAppearance = appearance
    navigationController?.navigationBar.shadowImage = UIImage()
  }

  private func setupSearchViews() {
    self.view.addSubview(topView)
    topView.topAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.topAnchor).isActive = true
    topView.leadingAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.leadingAnchor).isActive = true
    topView.trailingAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.trailingAnchor).isActive = true

    let topViewHeight: CGFloat = viewModel.addType == .calendar ? 56 + 32 + 10 : 56

    topView.heightAnchor.constraint(equalToConstant: topViewHeight).isActive = true
    topView.translatesAutoresizingMaskIntoConstraints = false
    topView.backgroundColor = .systemBackground
    topView.layer.shadowOffset = CGSize(width: 0, height: 1)
    topView.layer.shadowRadius = 0.5
    topView.layer.shadowOpacity = 0.5
    topView.layer.shadowColor = UIColor.gray.cgColor
    topView.layer.masksToBounds = false

    self.view.addSubview(searchBar)

    let addImage = setupMoreOptionsImage()
    addImage.image = UIImage(systemName: "plus")?.withRenderingMode(.alwaysTemplate)
    addImage.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -8).isActive = true
    addImage.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(addItem)))

    let sortImage = setupMoreOptionsImage()
    sortImage.image = UIImage(systemName: "arrow.up.arrow.down")?.withRenderingMode(.alwaysTemplate)
    sortImage.trailingAnchor.constraint(equalTo: addImage.leadingAnchor, constant: -8).isActive = true
    sortImage.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(showSortOptions)))

    searchBar.topAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.topAnchor).isActive = true
    searchBar.leadingAnchor.constraint(
      equalTo: self.view.safeAreaLayoutGuide.leadingAnchor,
      constant: 8).isActive = true
    searchBar.trailingAnchor.constraint(
      equalTo: sortImage.leadingAnchor,
      constant: -8).isActive = true
    searchBar.heightAnchor.constraint(equalToConstant: 56).isActive = true // default height since iOS 11
    searchBar.translatesAutoresizingMaskIntoConstraints = false
    searchBar.searchBarStyle = .minimal

    searchBar.placeholder = NSLocalizedString("Search...", comment: "")
    searchBar.backgroundColor = .clear
    searchBar.isTranslucent = false
    searchBar.delegate = self

    segmentedControl.selectedSegmentIndex = 0
    segmentedControl.backgroundColor = .clear

    if viewModel.addType == .calendar {
      self.view.addSubview(segmentedControl)
      segmentedControl.topAnchor.constraint(equalTo: searchBar.bottomAnchor).isActive = true
      segmentedControl.leadingAnchor.constraint(
        equalTo: view.safeAreaLayoutGuide.leadingAnchor,
        constant: 16).isActive = true
      segmentedControl.trailingAnchor.constraint(
        equalTo: view.safeAreaLayoutGuide.trailingAnchor,
        constant: -16).isActive = true
      segmentedControl.heightAnchor.constraint(equalToConstant: 32).isActive = true
      segmentedControl.translatesAutoresizingMaskIntoConstraints = false

      segmentedControl.addTarget(self, action: #selector(segmentatedControlDidChangeValue), for: .valueChanged)
    }
  }

  private func setupMoreOptionsImage() -> UIImageView {
    let imageView = UIImageView()
    imageView.tintColor = .label
    imageView.translatesAutoresizingMaskIntoConstraints = false
    imageView.contentMode = .scaleAspectFit
    imageView.isUserInteractionEnabled = true

    view.addSubview(imageView)
    imageView.heightAnchor.constraint(equalToConstant: 30).isActive = true
    imageView.widthAnchor.constraint(equalToConstant: 30).isActive = true
    imageView.centerYAnchor.constraint(equalTo: searchBar.centerYAnchor).isActive = true

    return imageView
  }

  // MARK: - Collection View
  private func setupContentViews() {
    view.addSubview(contentCollectionView)

    NSLayoutConstraint.activate([
      contentCollectionView.topAnchor.constraint(equalTo: topView.bottomAnchor, constant: 8),
      contentCollectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      contentCollectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
      contentCollectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])

    createDataSource()
    applyDataSource()
  }

  private func createLayout() -> UICollectionViewLayout {
    return UICollectionViewCompositionalLayout.listLayout(groupEstimatedHeight: 70)
  }

  private func createDataSource() {
    contentDataSource = ArticlesDataSource(
      collectionView: contentCollectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ArticleCell.reuseIdentifier,
        // swiftlint:disable:next force_cast
        for: indexPath) as! ArticleCell

      // swiftlint:disable force_unwrapping
      if let article = item as? Article {
        cell.config(with: article)
        cell.isChecked = self.viewModel.isSelected(id: article.id!)
      }

      if let dish = item as? Dish {
        cell.config(with: dish)
        cell.isChecked = self.viewModel.isSelected(id: dish.id!)
      }
      // swiftlint:enable force_unwrapping
      return cell
    }
  }

  private func applyDataSource() {
    var snapshot = ArticlesDataSnapshot()

    let section = Section(title: "", items: viewModel.filteredItems)

    snapshot.appendSections([section])
    snapshot.appendItems(section.items, toSection: section)

    contentDataSource.apply(snapshot, animatingDifferences: false)
  }

  // MARK: - Action handlers
  @objc private func hideKeyboard() {
    view.endEditing(true)
  }

  @objc private func addTapped() {
    viewModel.save()

    if viewModel.addType == .calendar {
      dismiss(animated: true)
    } else {
      navigationController?.popViewController(animated: true)
    }
  }

  @objc private func cancelTapped() {
    dismiss(animated: true)
  }

  @objc private func segmentatedControlDidChangeValue() {
    if segmentedControl.selectedSegmentIndex == 0 {
      viewModel.setSelectionType(to: .article)
    } else {
      viewModel.setSelectionType(to: .dish)
    }
  }

  @objc private func addItem() {
    switch viewModel.selectionType {
    case .article:
      let articleViewModel = NewArticleViewModel()
      let newController = NewArticleViewController(withViewModel: articleViewModel)
      let navController = UINavigationController(rootViewController: newController)
      present(navController, animated: true)
    case .dish:
      let articleViewModel = NewDishViewModel()
      let newController = NewDishViewController(withViewModel: articleViewModel)
      let navController = UINavigationController(rootViewController: newController)
      present(navController, animated: true)
    }
  }

  @objc private func showSortOptions() {
    let sortAlert = UIAlertController(
      title: NSLocalizedString("Sort by...", comment: ""),
      message: nil,
      preferredStyle: .actionSheet)

    sortAlert.addAction(UIAlertAction(
      title: NSLocalizedString("Alphabetically", comment: ""),
      style: .default) { [weak self] _ in
      self?.viewModel.setSort(by: .alphabetically)
    })

    sortAlert.addAction(UIAlertAction(
      title: NSLocalizedString("Color", comment: ""),
      style: .default) { [weak self] _ in
      self?.viewModel.setSort(by: .color)
    })

    sortAlert.addAction(UIAlertAction(title: NSLocalizedString("Cancel", comment: ""), style: .cancel))
    sortAlert.popoverPresentationController?.barButtonItem = self.navigationItem.rightBarButtonItem
    present(sortAlert, animated: true)
  }
}

extension NewItemCalendarViewController: UISearchBarDelegate {
  func searchBar(_ searchBar: UISearchBar, textDidChange searchText: String) {
    viewModel.filter(withText: searchText)
  }
}

extension NewItemCalendarViewController: UICollectionViewDelegate {
  func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
    let cell = collectionView.dequeueReusableCell(
      withReuseIdentifier: ArticleCell.reuseIdentifier,
      // swiftlint:disable force_cast
      for: indexPath) as! ArticleCell

    let itemSelected = contentDataSource.itemIdentifier(for: indexPath)

    // swiftlint:disable force_unwrapping
    if let article = itemSelected as? Article {
      viewModel.didSelect(id: article.id!)
    }

    if let dish = itemSelected as? Dish {
      viewModel.didSelect(id: dish.id!)
    }

    // swiftlint:enable force_unwrapping
    cell.isChecked.toggle()
  }
}
