//
//  ColorsItemSelector.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 23/4/21.
//

import UIKit

private typealias DataSource = UICollectionViewDiffableDataSource<ColorSection, ColorItem>
private typealias DataSnapshot = NSDiffableDataSourceSnapshot<ColorSection, ColorItem>

class ColorsItemSelector: UIView {
  private let colorLabel: UILabel = {
    let lbl = UILabel()
    lbl.font = .newYorkFont(weight: .regular, andTextStyle: .headline)
    lbl.textColor = .systemGray
    lbl.translatesAutoresizingMaskIntoConstraints = false
    lbl.text = NSLocalizedString("Color", comment: "")
    return lbl
  }()

  private lazy var colorsCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createColorsCollectionLayout())
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ColorsItemSelectorCell.self, forCellWithReuseIdentifier: ColorsItemSelectorCell.reuseIdentifier)
    cvw.delegate = self
    return cvw
  }()

  // swiftlint:disable:next implicitly_unwrapped_optional
  private var dataSource: DataSource!
  // swiftlint:disable:next implicitly_unwrapped_optional
  private var section: ColorSection!
  private let availableColors: [UIColor] = [.appGreen, .appRed, .appPurple, .systemOrange, .systemYellow, .systemBlue]

  var selectedColor: UIColor {
    section.items.first { $0.isSelected }?.color ?? .appGreen
  }

  override init(frame: CGRect) {
    super.init(frame: frame)
    setupViews()
    initOptions()
    createDataSource()
    applyDataSource()
  }

  required init?(coder: NSCoder) {
    fatalError("Not happening")
  }

  private func setupViews() {
    addSubview(colorLabel)
    addSubview(colorsCollectionView)

    NSLayoutConstraint.activate([
      colorLabel.topAnchor.constraint(equalTo: topAnchor),
      colorLabel.leadingAnchor.constraint(equalTo: leadingAnchor),
      colorLabel.trailingAnchor.constraint(equalTo: trailingAnchor),

      colorsCollectionView.topAnchor.constraint(equalTo: colorLabel.bottomAnchor, constant: 8),
      colorsCollectionView.leadingAnchor.constraint(equalTo: leadingAnchor),
      colorsCollectionView.trailingAnchor.constraint(equalTo: trailingAnchor),
      colorsCollectionView.heightAnchor.constraint(equalToConstant: 44)
    ])
  }

  private func initOptions() {
    var items = availableColors.map { ColorItem(color: $0) }
    items[0].isSelected = true
    section = ColorSection(items: items)
  }

  private func createColorsCollectionLayout() -> UICollectionViewLayout {
    let itemSize = NSCollectionLayoutSize(
      widthDimension: .fractionalWidth(1.0),
      heightDimension: .fractionalHeight(1.0))
    let layoutItem = NSCollectionLayoutItem(layoutSize: itemSize)

    let layoutGroupSize = NSCollectionLayoutSize(
      widthDimension: .estimated(40),
      heightDimension: .estimated(40))

    let layoutGroup = NSCollectionLayoutGroup.horizontal(layoutSize: layoutGroupSize, subitems: [layoutItem])

    let section = NSCollectionLayoutSection(group: layoutGroup)
    section.orthogonalScrollingBehavior = .continuous
    section.interGroupSpacing = 12
    return UICollectionViewCompositionalLayout(section: section)
  }

  private func createDataSource() {
    dataSource = DataSource(
      collectionView: colorsCollectionView) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ColorsItemSelectorCell.reuseIdentifier,
        // swiftlint:disable:next force_cast
        for: indexPath) as! ColorsItemSelectorCell

      cell.config(forItem: item)
      return cell
    }
  }

  private func applyDataSource() {
    var snapshot = DataSnapshot()
    snapshot.appendSections([section])
    snapshot.appendItems(section.items, toSection: section)
    dataSource.apply(snapshot, animatingDifferences: false)
  }
}

extension ColorsItemSelector: UICollectionViewDelegate {
  func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
    setSelected(color: section.items[indexPath.item].color)
  }

  func setSelected(color: UIColor) {
    for index in section.items.indices {
      section.items[index].isSelected = false
    }


    if let selectedColor = section.items.first(where: { $0.color == color }),
      let index = section.items.firstIndex(of: selectedColor) {
      section.items[index].isSelected = true
    } else {
      section.items[0].isSelected = true
    }

    applyDataSource()
  }
}
