//
//  NewDishViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 24/4/21.
//

import UIKit

class NewDishViewController: UIViewController {
  // swiftlint:disable implicitly_unwrapped_optional
  private var nameTextView: UITextView!
  private var commentTextView: UITextView!
  private let colorsItemSelector = ColorsItemSelector()
  private var articlesDataSource: ArticlesDataSource!
  // swiftlint:enable implicitly_unwrapped_optional

  private lazy var articlesCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createArticlesLayout())
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ArticleCell.self, forCellWithReuseIdentifier: ArticleCell.reuseIdentifier)
    cvw.register(
      TitleAndAddHeaderView.self,
      forSupplementaryViewOfKind: UICollectionView.elementKindSectionHeader,
      withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier)
    cvw.delegate = self
    return cvw
  }()

  private var viewModel: NewDishViewModel

  deinit {
    viewModel.discardDish()
  }

  init(withViewModel: NewDishViewModel) {
    viewModel = withViewModel
    super.init(nibName: nil, bundle: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("nope")
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    applyArticlesDataSource()
  }

  override func viewDidLoad() {
    super.viewDidLoad()

    view.backgroundColor = .systemBackground

    let controllerTitle = viewModel.isNew ? "New dish" : "Dish"
    navigationItem.title = NSLocalizedString(controllerTitle, comment: "")

    let rightActionTitle = viewModel.isNew ? "Add" : "Update"

    navigationItem.rightBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString(rightActionTitle, comment: ""),
      style: .plain,
      target: self,
      action: #selector(addTapped))

    navigationItem.rightBarButtonItem?.tintColor = .label

    navigationItem.leftBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString("Cancel", comment: ""),
      style: .plain,
      target: self,
      action: #selector(cancelTapped))

    navigationItem.leftBarButtonItem?.tintColor = .label

    let tap = UITapGestureRecognizer(target: self, action: #selector(hideKeyboard))
    tap.cancelsTouchesInView = false
    view.addGestureRecognizer(tap)

    setupViews()
    setupArticlesDataSource()

    setupContent()

    viewModel.dishDidUpdate = { [weak self] in
      self?.setupContent()
    }
  }

  private func setupViews() {
    nameTextView = createNewTextView(withText: "Give a name to your dish")
    nameTextView.textContainer.maximumNumberOfLines = 2
    commentTextView = createNewTextView(withText: "Take notes about your dish")

    let nameLabel = createNewSectionTitleLabel()

    nameLabel.text = NSLocalizedString("Name", comment: "")

    let commentLabel = createNewSectionTitleLabel()
    commentLabel.text = NSLocalizedString("Comment", comment: "")

    view.addSubview(nameLabel)
    view.addSubview(nameTextView)
    view.addSubview(commentLabel)
    view.addSubview(commentTextView)
    view.addSubview(colorsItemSelector)
    view.addSubview(articlesCollectionView)

    colorsItemSelector.translatesAutoresizingMaskIntoConstraints = false

    NSLayoutConstraint.activate([
      nameLabel.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor, constant: 16),
      nameLabel.leadingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leadingAnchor, constant: 16),
      nameLabel.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -16),

      nameTextView.topAnchor.constraint(equalTo: nameLabel.bottomAnchor, constant: 6),
      nameTextView.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      nameTextView.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
      nameTextView.heightAnchor.constraint(equalToConstant: 44),

      commentLabel.topAnchor.constraint(equalTo: nameTextView.bottomAnchor, constant: 16),
      commentLabel.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      commentLabel.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),

      commentTextView.topAnchor.constraint(equalTo: commentLabel.bottomAnchor, constant: 8),
      commentTextView.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      commentTextView.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
      commentTextView.heightAnchor.constraint(equalToConstant: 150),

      colorsItemSelector.topAnchor.constraint(equalTo: commentTextView.bottomAnchor, constant: 16),
      colorsItemSelector.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
      colorsItemSelector.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
      colorsItemSelector.heightAnchor.constraint(equalToConstant: 75),

      articlesCollectionView.topAnchor.constraint(equalTo: colorsItemSelector.bottomAnchor, constant: 8),
      articlesCollectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      articlesCollectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
      articlesCollectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])
  }

  private func setupContent() {
    if let dish = viewModel.currentDish {
      if let name = dish.name {
        nameTextView.text = name
        nameTextView.textColor = .label
      }

      if let comment = dish.comment, !comment.isEmpty {
        commentTextView.textColor = .label
        commentTextView.text = comment
      }

      colorsItemSelector.setSelected(color: dish.wrappedColor)
    }

    applyArticlesDataSource()
  }

  private func createNewSectionTitleLabel() -> UILabel {
    let lbl = UILabel()
    lbl.font = .newYorkFont(weight: .regular, andTextStyle: .headline)
    lbl.textColor = .systemGray
    lbl.translatesAutoresizingMaskIntoConstraints = false
    return lbl
  }

  private func createNewTextView(withText text: String) -> UITextView {
    let textView = UITextView()
    textView.layer.cornerRadius = 15
    textView.layer.masksToBounds = true
    textView.translatesAutoresizingMaskIntoConstraints = false
    textView.textColor = .systemGray
    textView.backgroundColor = .systemGray5
    textView.textContainerInset = UIEdgeInsets(top: 12, left: 10, bottom: 12, right: 10)
    textView.font = .newYorkFont(weight: .regular, andTextStyle: .body)
    textView.delegate = self
    textView.text = NSLocalizedString(text, comment: "")
    return textView
  }

  // MARK: - Action handlers
  @objc private func addTapped() {
    guard nameTextView.textColor == .label else {
      showAlertError(withMessage: NSLocalizedString("The article must have a name", comment: ""))
      return
    }

    viewModel.saveNewDish(
      withName: nameTextView.text,
      comment: commentTextView.textColor == .label ? commentTextView.text : "",
      andColor: colorsItemSelector.selectedColor) { [weak self] error in
      switch error {
      case .none:
        self?.dismiss(animated: true)
      case .name:
        self?.showAlertError(withMessage: NSLocalizedString("There's already an article with this name", comment: ""))
      case .noname:
        self?.showAlertError(withMessage: NSLocalizedString("The article must have a name", comment: ""))
      case .coreData:
        self?.showAlertError(withMessage: NSLocalizedString("Error storing the article", comment: ""))
      }
    }
  }

  @objc private func cancelTapped() {
    dismiss(animated: true)
  }

  @objc private func hideKeyboard() {
    view.endEditing(true)
  }

  // MARK: - CollectionView
  private func createArticlesLayout() -> UICollectionViewLayout {
    return UICollectionViewCompositionalLayout.listWithHeaderLayout(groupEstimatedHeight: 70)
  }

  private func setupArticlesDataSource() {
    // swiftlint:disable force_cast
    articlesDataSource = ArticlesDataSource(
      collectionView: articlesCollectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ArticleCell.reuseIdentifier,
        for: indexPath) as! ArticleCell

      if let article = item as? Article {
        cell.config(with: article)
      }

      return cell
    }

    articlesDataSource.supplementaryViewProvider = { [weak self] collectionView, kind, indexPath in
      switch kind {
      case UICollectionView.elementKindSectionHeader:
        let header = collectionView.dequeueReusableSupplementaryView(
          ofKind: kind,
          withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier,
          for: indexPath) as! TitleAndAddHeaderView

        if let section = self?.articlesDataSource?.snapshot().sectionIdentifiers[indexPath.section] {
          header.configure(with: section.title)
          header.addAction = { [weak self] in
            let calendarViewModel = NewItemCalendarViewModel(withDish: self?.viewModel.currentDish)
            self?.navigationController?.pushViewController(
              NewItemCalendarViewController(withViewModel: calendarViewModel),
              animated: true)
          }
          return header
        }

      default:
        fatalError("Implement new kind")
      }

      return nil
    }
    // swiftlint:enable force:cast
  }

  private func applyArticlesDataSource() {
    var snapshot = ArticlesDataSnapshot()
    // swiftlint:disable:next force_unwrapping
    let sortedArray = viewModel.currentDish?.articlesArray.sorted { $0.name!.lowercased() < $1.name!.lowercased() }

    let section = Section(
      title: NSLocalizedString("Articles", comment: ""),
      items: sortedArray ?? [])

    snapshot.appendSections([section])
    snapshot.appendItems(section.items, toSection: section)

    articlesDataSource.apply(snapshot)
  }
}

extension NewDishViewController: UICollectionViewDelegate {
  func collectionView(
    _ collectionView: UICollectionView,
    contextMenuConfigurationForItemAt indexPath: IndexPath,
    point: CGPoint
  ) -> UIContextMenuConfiguration? {
    UIContextMenuConfiguration(
      identifier: nil,
      previewProvider: nil) {  _ in
      let deleteAction = self.deleteItemAction(at: indexPath)
      let children = [deleteAction]

      return UIMenu(title: "", image: nil, children: children)
    }
  }

  private func deleteItemAction(at indexPath: IndexPath) -> UIAction {
    let image = UIImage(systemName: "trash")

    return UIAction(
      title: NSLocalizedString("Delete", comment: ""),
      image: image,
      attributes: .destructive) { [weak self] _ in
      self?.showConfirmationAlert(
        withMessage: NSLocalizedString("Do you want to delete this article?", comment: "")
      ) { [weak self] in
        guard let self = self else { return }
        let selectedArticle = self.articlesDataSource.itemIdentifier(for: indexPath) as! Article
        self.viewModel.delete(article: selectedArticle)
        self.applyArticlesDataSource()
      }
    }
  }
}

extension NewDishViewController: UITextViewDelegate {
  func textViewDidBeginEditing(_ textView: UITextView) {
    if textView.textColor != .label {
      textView.text = ""
      textView.textColor = .label
    }
  }
}
