//
//  ListsViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 3/4/21.
//

import UIKit

class ListsViewController: UIViewController {
  // swiftlint:disable:next convenience_type
  private struct RightItemConstraint {
    static let imageSizeForLargeState: CGFloat = 36
    static let imageRightMargin: CGFloat = 16
    static let imageBottomMarginForLargeState: CGFloat = 16
    static let imageBottomMarginForSmallState: CGFloat = 12
    static let imageSizeForSmallState: CGFloat = 32
    static let navBarHeightSmallState: CGFloat = 44
    static let navBarHeightLargeState: CGFloat = 96.5
  }

  // swiftlint:disable implicitly_unwrapped_optional
  private var collectionView: UICollectionView!
  private var dataSource: ArticlesDataSource!
  var viewModel: SimpleListViewModelProtocol!
  // swiftlint:enable implicitly_unwrapped_optional

  override func viewDidLoad() {
    super.viewDidLoad()
    view.backgroundColor = .systemBackground

    guard viewModel != nil else {
      fatalError("View model not initialized")
    }

    collectionView = UICollectionView(frame: .zero, collectionViewLayout: createLayout())
    collectionView.delegate = self

    view.addSubview(collectionView)

    collectionView.contentInset.top = 10
    collectionView.backgroundColor = .clear
    collectionView.translatesAutoresizingMaskIntoConstraints = false

    NSLayoutConstraint.activate([
      collectionView.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor),
      collectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      collectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
      collectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])

    registerCells()
    createDataSource()

    viewModel.itemsDidChange = { items in
      DispatchQueue.main.async { [weak self] in
        self?.applyDataSource(withItems: items)
      }
    }
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    tabBarController?.navigationItem.title = viewModel.navigationItemTitle
    tabBarController?.navigationController?.navigationBar.prefersLargeTitles = true
    createRightNavButton()
    viewModel.fetch()
  }

  private func registerCells() {
    switch viewModel.type {
    case .lists:
      collectionView.register(ShoppingListCell.self, forCellWithReuseIdentifier: ShoppingListCell.reuseIdentifier)
    case .dishes:
      collectionView.register(ShoppingListCell.self, forCellWithReuseIdentifier: ShoppingListCell.reuseIdentifier)
      collectionView.register(
        SubtitleHeaderView.self,
        forSupplementaryViewOfKind: UICollectionView.elementKindSectionHeader,
        withReuseIdentifier: SubtitleHeaderView.reuseIdentifier)
    default:
      collectionView.register(ArticleCell.self, forCellWithReuseIdentifier: ArticleCell.reuseIdentifier)
    }
  }

  private func createLayout() -> UICollectionViewLayout {
    switch viewModel.type {
    case .dishes:
      return UICollectionViewCompositionalLayout.listWithHeaderLayout(
        groupEstimatedHeight: viewModel.groupHeight,
        isStickyHeader: false)
    default:
      return UICollectionViewCompositionalLayout.listLayout(groupEstimatedHeight: viewModel.groupHeight)
    }
  }

  // swiftlint:disable:next cyclomatic_complexity
  private func createDataSource() {
    dataSource = ArticlesDataSource(
      collectionView: collectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      switch self.viewModel.type {
      case .lists:
        if let cell = collectionView.dequeueReusableCell(
          withReuseIdentifier: ShoppingListCell.reuseIdentifier,
          for: indexPath) as? ShoppingListCell,
          let shoppingList = item as? ShoppingList {
          cell.config(with: shoppingList)
          return cell
        }
      case .dishes:
        if let cell = collectionView.dequeueReusableCell(
          withReuseIdentifier: ShoppingListCell.reuseIdentifier,
          for: indexPath) as? ShoppingListCell,
          let dish = item as? Dish {
          cell.config(with: dish)
          return cell
        }
      case .articles:
        if let cell = collectionView.dequeueReusableCell(
          withReuseIdentifier: ArticleCell.reuseIdentifier,
          for: indexPath) as? ArticleCell,
          let article = item as? Article {
          cell.config(with: article)
          return cell
        }
      }

      fatalError("type not configured")
    }

    if viewModel.type == .dishes {
      dataSource.supplementaryViewProvider = { [weak self] collectionView, kind, indexPath in
        guard let self = self else { return nil }

        switch kind {
        case UICollectionView.elementKindSectionHeader:

          let header = collectionView.dequeueReusableSupplementaryView(
            ofKind: kind,
            withReuseIdentifier: SubtitleHeaderView.reuseIdentifier,
            // swiftlint:disable:next force_cast
            for: indexPath) as! SubtitleHeaderView

          if let section = self.dataSource?.snapshot().sectionIdentifiers[indexPath.section] {
            header.configure(with: section.title)
            return header
          }

        default:
          fatalError("Implement new kind")
        }

        return nil
      }
    }
  }

  private func applyDataSource(withItems items: [AnyHashable]) {
    // fake data to test view
    var snapshot = ArticlesDataSnapshot()

    switch viewModel.type {
    case .lists:
      let section = Section(items: viewModel.items)
      snapshot.appendSections([section])
      snapshot.appendItems(section.items, toSection: section)
    case .dishes:
      // swiftlint:disable:next force_cast
      let vmdl = viewModel as! DishesViewModel
      let sections = vmdl.getSections()
      snapshot.appendSections(sections)
      sections.forEach { section in
        snapshot.appendItems(section.items, toSection: section)
      }
    case .articles:
      let section = Section(items: items)
      snapshot.appendSections([section])
      snapshot.appendItems(section.items, toSection: section)
    }

    dataSource.apply(snapshot, animatingDifferences: false)
  }

  // MARK: - RightBarButton here. First tab
  private func createRightNavButton() {
    // swiftlint:disable force_unwrapping
    guard let tabController = tabBarController as? LandingViewController else { return }
    guard let navigationBar = tabBarController?.navigationController?.navigationBar else { return }
    // if it's not nil, it's already placed
    guard tabController.rightButtonImageView == nil else {
      tabController.rightButtonImageView?.image = UIImage(systemName: "plus")
      tabController.rightButtonImageView?.gestureRecognizers?.removeAll()
      tabController.rightButtonImageView?.addGestureRecognizer(
        UITapGestureRecognizer(target: self, action: #selector(showNewController))
      )
      tabController.rightButtonImageView?.isHidden = false
      return
    }

    tabController.rightButtonImageView = UIImageView(image: UIImage(systemName: "plus"))
    tabController.rightButtonImageView!.tintColor = .label

    navigationBar.addSubview(tabController.rightButtonImageView!)
    tabController.rightButtonImageView!.translatesAutoresizingMaskIntoConstraints = false
    NSLayoutConstraint.activate([
      tabController.rightButtonImageView!.rightAnchor.constraint(
        equalTo: navigationBar.rightAnchor,
        constant: -RightItemConstraint.imageRightMargin),
      tabController.rightButtonImageView!.bottomAnchor.constraint(
        equalTo: navigationBar.bottomAnchor,
        constant: -RightItemConstraint.imageBottomMarginForSmallState),
      tabController.rightButtonImageView!.heightAnchor.constraint(
        equalToConstant: RightItemConstraint.imageSizeForSmallState),
      tabController.rightButtonImageView!.widthAnchor.constraint(
        equalTo: tabController.rightButtonImageView!.heightAnchor)
    ])

    // swiftlint:enable force_unwrapping
    tabController.rightButtonImageView?.isUserInteractionEnabled = true
    tabController.rightButtonImageView?.addGestureRecognizer(
      UITapGestureRecognizer(target: self, action: #selector(showNewController))
    )

    tabController.rightButtonImageView?.isHidden = false
  }

  @objc private func showNewController() {
    let navController: UINavigationController
    switch viewModel.type {
    case .lists:
      let newController = NewListViewController(withViewModel: NewShoppingListViewModel())
      navController = UINavigationController(rootViewController: newController)
    case .dishes:
      let newController = NewDishViewController(withViewModel: NewDishViewModel())
      navController = UINavigationController(rootViewController: newController)
    case .articles:
      let newController = NewArticleViewController(withViewModel: NewArticleViewModel())
      navController = UINavigationController(rootViewController: newController)
    }

    present(navController, animated: true)
  }
}

extension ListsViewController: UICollectionViewDelegate, UICollectionViewDelegateFlowLayout {
  func scrollViewDidScroll(_ scrollView: UIScrollView) {
    guard let height = navigationController?.navigationBar.frame.height else { return }
    moveAndResizeImage(for: height)
  }

  private func moveAndResizeImage(for height: CGFloat) {
    guard let tabController = tabBarController as? LandingViewController else { return }

    let coeff: CGFloat = {
      let delta = height - RightItemConstraint.navBarHeightSmallState
      let heightDifferenceBetweenStates =
        RightItemConstraint.navBarHeightLargeState - RightItemConstraint.navBarHeightSmallState

      return delta / heightDifferenceBetweenStates
    }()

    let factor = RightItemConstraint.imageSizeForSmallState / RightItemConstraint.imageSizeForLargeState

    let scale: CGFloat = {
      let sizeAddendumFactor = coeff * (1.0 - factor)
      return min(1.0, sizeAddendumFactor + factor)
    }()

    // Value of difference between icons for large and small states
    let sizeDiff = RightItemConstraint.imageSizeForLargeState * (1.0 - factor) // 8.0
    let yTranslation: CGFloat = {
      /// This value = 14. It equals to difference of 12 and 6 (bottom margin for large and small states). Also it adds 8.0 (size difference when the image gets smaller size)
      let maxYTranslation = RightItemConstraint.imageBottomMarginForLargeState -
        RightItemConstraint.imageBottomMarginForSmallState + sizeDiff
      return max(
        0,
        min(
          maxYTranslation,
          (maxYTranslation - coeff * (RightItemConstraint.imageBottomMarginForSmallState + sizeDiff))))
    }()

    let xTranslation = max(0, sizeDiff - coeff * sizeDiff)

    tabController.rightButtonImageView?.transform = CGAffineTransform.identity
      .scaledBy(x: scale, y: scale)
      .translatedBy(x: xTranslation, y: yTranslation)
  }

  func collectionView(
    _ collectionView: UICollectionView,
    contextMenuConfigurationForItemAt indexPath: IndexPath,
    point: CGPoint
  ) -> UIContextMenuConfiguration? {
    UIContextMenuConfiguration(
      identifier: nil,
      previewProvider: nil) {  _ in
      let deleteAction = self.deleteItemAction(at: indexPath)

      let children: [UIAction]

      if self.viewModel.type == .lists {
        let purchase = self.purchaseListAction(at: indexPath)
        children = [purchase, deleteAction]
      } else {
        children = [deleteAction]
      }

      return UIMenu(title: "", image: nil, children: children)
    }
  }

  private func deleteItemAction(at indexPath: IndexPath) -> UIAction {
    let image = UIImage(systemName: "trash")

    return UIAction(
      title: NSLocalizedString("Delete", comment: ""),
      image: image,
      attributes: .destructive) { [weak self] _ in
      self?.showConfirmationAlert(withMessage: "Do you want to delete this article?") { [weak self] in
        guard let self = self else { return }
        self.viewModel.delete(self.viewModel.items[indexPath.item])
      }
    }
  }

  private func purchaseListAction(at indexPath: IndexPath) -> UIAction {
    let image = UIImage(systemName: "cart")

    return UIAction(
      title: NSLocalizedString("Purchase", comment: ""),
      image: image) { [weak self] _ in
      guard let self = self else { return }
      let itemSelected = self.viewModel.items[indexPath.item]

      if let list = itemSelected as? ShoppingList {
        let newViewModel = ShopViewModel(withList: list)
        let viewController = ShopViewController(withViewModel: newViewModel)
        self.navigationController?.pushViewController(viewController, animated: true)

        guard let tabController = self.tabBarController as? LandingViewController else { return }
        tabController.rightButtonImageView?.isHidden = true
      }
    }
  }

  func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
    let itemSelected = viewModel.items[indexPath.item]

    switch viewModel.type {
    case .lists:
      if let list = itemSelected as? ShoppingList {
        let viewModel = NewShoppingListViewModel(withList: list)
        let newController = NewListViewController(withViewModel: viewModel)
        let navController = UINavigationController(rootViewController: newController)
        present(navController, animated: true)
      }
    case .dishes:
      if let dish = itemSelected as? Dish {
        let dishViewModel = NewDishViewModel(withDish: dish)
        let newController = NewDishViewController(withViewModel: dishViewModel)
        let navController = UINavigationController(rootViewController: newController)
        present(navController, animated: true)
      }
    case .articles:
      if let article = itemSelected as? Article {
        let articleViewModel = NewArticleViewModel(withArticle: article)
        let newController = NewArticleViewController(withViewModel: articleViewModel)
        let navController = UINavigationController(rootViewController: newController)
        present(navController, animated: true)
      }
    }
  }
}
