//
//  NewListViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 25/4/21.
//

import UIKit

class NewListViewController: UIViewController {
  private let nameLabel: UILabel = {
    let lbl = UILabel()
    lbl.font = .newYorkFont(weight: .regular, andTextStyle: .headline)
    lbl.textColor = .systemGray
    lbl.translatesAutoresizingMaskIntoConstraints = false
    lbl.text = NSLocalizedString("Name", comment: "")
    return lbl
  }()

  private lazy var nameTextView: UITextView = {
    let textView = UITextView()
    textView.layer.cornerRadius = 15
    textView.layer.masksToBounds = true
    textView.translatesAutoresizingMaskIntoConstraints = false
    textView.textColor = .systemGray
    textView.backgroundColor = .systemGray5
    textView.textContainerInset = UIEdgeInsets(top: 12, left: 10, bottom: 12, right: 10)
    textView.font = .newYorkFont(weight: .regular, andTextStyle: .body)
    textView.textContainer.maximumNumberOfLines = 2
    textView.delegate = self
    textView.text = NSLocalizedString("Give a name to your list", comment: "")
    return textView
  }()

  private lazy var articlesCollectionView: UICollectionView = {
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: createArticlesLayout())
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ArticleCell.self, forCellWithReuseIdentifier: ArticleCell.reuseIdentifier)
    cvw.register(
      TitleAndAddHeaderView.self,
      forSupplementaryViewOfKind: UICollectionView.elementKindSectionHeader,
      withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier)
    cvw.delegate = self
    return cvw
  }()

  private lazy var cartImage: UIImageView = {
    let iv = UIImageView()
    iv.image = UIImage(systemName: "cart")?.withRenderingMode(.alwaysTemplate)
    iv.tintColor = .label
    iv.contentMode = .scaleAspectFit
    iv.translatesAutoresizingMaskIntoConstraints = false
    iv.backgroundColor = .clear
    iv.isUserInteractionEnabled = true
    iv.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(shop)))
    return iv
  }()

  // swiftlint:disable:next implicitly_unwrapped_optional
  private var articlesDataSource: ArticlesDataSource!

  private var viewModel: NewShoppingListViewModel

  deinit {
    viewModel.discardList()
  }

  init(withViewModel: NewShoppingListViewModel) {
    viewModel = withViewModel
    super.init(nibName: nil, bundle: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("nope")
  }

  override func viewDidLoad() {
    super.viewDidLoad()
    view.backgroundColor = .systemBackground

    let controllerTitle = viewModel.isNew ? "New List" : "List"
    navigationItem.title = NSLocalizedString(controllerTitle, comment: "")

    let rightActionTitle = viewModel.isNew ? "Add" : "Update"

    navigationItem.rightBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString(rightActionTitle, comment: ""),
      style: .plain,
      target: self,
      action: #selector(addTapped))

    navigationItem.rightBarButtonItem?.tintColor = .label

    navigationItem.leftBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString("Cancel", comment: ""),
      style: .plain,
      target: self,
      action: #selector(cancelTapped))

    navigationItem.leftBarButtonItem?.tintColor = .label

    let tap = UITapGestureRecognizer(target: self, action: #selector(hideKeyboard))
    tap.cancelsTouchesInView = false
    view.addGestureRecognizer(tap)

    setupViews()
    setupArticlesDataSource()

    setupContent()

    viewModel.listDidUpdate = { [weak self] in
      self?.setupContent()
    }
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    applyArticlesDataSource()
  }

  private func setupViews() {
    view.addSubview(nameLabel)
    view.addSubview(nameTextView)
    view.addSubview(articlesCollectionView)

    NSLayoutConstraint.activate([
      nameLabel.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor, constant: 16),
      nameLabel.leadingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leadingAnchor, constant: 16),
      nameLabel.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -16)
    ])

    if !viewModel.currentList.itemsArray.isEmpty {
      view.addSubview(cartImage)
      NSLayoutConstraint.activate([
        cartImage.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -16),
        cartImage.centerYAnchor.constraint(equalTo: nameTextView.centerYAnchor),
        cartImage.widthAnchor.constraint(equalToConstant: 30),
        cartImage.heightAnchor.constraint(equalToConstant: 30),

        nameTextView.topAnchor.constraint(equalTo: nameLabel.bottomAnchor, constant: 6),
        nameTextView.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
        nameTextView.trailingAnchor.constraint(equalTo: cartImage.leadingAnchor, constant: -16),
        nameTextView.heightAnchor.constraint(equalToConstant: 44)
      ])
    } else {
      NSLayoutConstraint.activate([
        nameTextView.topAnchor.constraint(equalTo: nameLabel.bottomAnchor, constant: 6),
        nameTextView.leadingAnchor.constraint(equalTo: nameLabel.leadingAnchor),
        nameTextView.trailingAnchor.constraint(equalTo: nameLabel.trailingAnchor),
        nameTextView.heightAnchor.constraint(equalToConstant: 44)
      ])
    }

    NSLayoutConstraint.activate([
      articlesCollectionView.topAnchor.constraint(equalTo: nameTextView.bottomAnchor, constant: 8),
      articlesCollectionView.leadingAnchor.constraint(equalTo: view.leadingAnchor),
      articlesCollectionView.trailingAnchor.constraint(equalTo: view.trailingAnchor),
      articlesCollectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])
  }

  private func setupContent() {
    if let name = viewModel.currentList.name, !name.isEmpty {
      nameTextView.text = name
      nameTextView.textColor = .label
    }

    applyArticlesDataSource()
  }

  private func setupArticlesDataSource() {
    // swiftlint:disable force_cast
    articlesDataSource = ArticlesDataSource(
      collectionView: articlesCollectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ArticleCell.reuseIdentifier,
        for: indexPath) as! ArticleCell
      if let articleItem = item as? ShoppingListItem,
        let article = articleItem.article {
        cell.config(with: article)
      }
      return cell
    }

    articlesDataSource.supplementaryViewProvider = { [weak self] collectionView, kind, indexPath in
      switch kind {
      case UICollectionView.elementKindSectionHeader:
        let header = collectionView.dequeueReusableSupplementaryView(
          ofKind: kind,
          withReuseIdentifier: TitleAndAddHeaderView.reuseIdentifier,
          for: indexPath) as! TitleAndAddHeaderView

        if let section = self?.articlesDataSource.snapshot().sectionIdentifiers[indexPath.section] {
          header.configure(with: section.title)
          header.addAction = { [weak self] in
            guard let self = self else { return }

            let calendarViewModel = NewItemCalendarViewModel(withList: self.viewModel.currentList)
            self.navigationController?.pushViewController(
              NewItemCalendarViewController(withViewModel: calendarViewModel),
              animated: true)
          }

          return header
        }

      default:
        fatalError("Implement new kind")
      }

      return nil
    }
    // swiftlint:enable force:cast
  }

  private func applyArticlesDataSource() {
    var snapshot = ArticlesDataSnapshot()

    let listItems = viewModel.currentList.itemsArray.filter {
      $0.article != nil && !$0.didDiscardItem
    }

    let section = Section(
      title: NSLocalizedString("Articles", comment: ""),
      items: listItems)

    snapshot.appendSections([section])
    snapshot.appendItems(section.items, toSection: section)

    articlesDataSource.apply(snapshot)
  }

  @objc private func hideKeyboard() {
    view.endEditing(true)
  }

  @objc private func addTapped() {
    guard nameTextView.textColor == .label else {
      showAlertError(withMessage: NSLocalizedString("The article must have a name", comment: ""))
      return
    }

    viewModel.save(withName: nameTextView.text) { [weak self] error in
      switch error {
      case .none:
        self?.dismiss(animated: true)
      case .name:
        return
      case .noname:
        self?.showAlertError(withMessage: NSLocalizedString("The list must have a name", comment: ""))
      case .coreData:
        self?.showAlertError(withMessage: NSLocalizedString("Error storing the list", comment: ""))
      }
    }
  }

  @objc private func cancelTapped() {
    dismiss(animated: true)
  }

  @objc private func shop() {
    let viewModel = ShopViewModel(withList: viewModel.currentList)
    let viewController = ShopViewController(withViewModel: viewModel)
    navigationController?.pushViewController(viewController, animated: true)
  }

  private func createArticlesLayout() -> UICollectionViewLayout {
    return UICollectionViewCompositionalLayout.listWithHeaderLayout(groupEstimatedHeight: 70)
  }
}

extension NewListViewController: UITextViewDelegate {
  func textViewDidBeginEditing(_ textView: UITextView) {
    textView.text = ""
    textView.textColor = .label
  }
}

extension NewListViewController: UICollectionViewDelegate {
  func collectionView(
    _ collectionView: UICollectionView,
    contextMenuConfigurationForItemAt indexPath: IndexPath,
    point: CGPoint
  ) -> UIContextMenuConfiguration? {
    UIContextMenuConfiguration(
      identifier: nil,
      previewProvider: nil) {  _ in
      let deleteAction = self.deleteItemAction(at: indexPath)
      let children = [deleteAction]

      return UIMenu(title: "", image: nil, children: children)
    }
  }

  private func deleteItemAction(at indexPath: IndexPath) -> UIAction {
    let image = UIImage(systemName: "trash")

    return UIAction(
      title: NSLocalizedString("Delete", comment: ""),
      image: image,
      attributes: .destructive) { [weak self] _ in
      self?.showConfirmationAlert(
        withMessage: NSLocalizedString("Do you want to delete this article?", comment: "")
      ) { [weak self] in
        guard let self = self else { return }
        let selectedItem = self.articlesDataSource.itemIdentifier(for: indexPath) as! ShoppingListItem
        self.viewModel.delete(item: selectedItem)
        self.applyArticlesDataSource()
      }
    }
  }
}
