//
//  ShopViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 9/5/21.
//

import UIKit

class ShopViewController: UIViewController {
  private lazy var articlesCollectionView: UICollectionView = {
    let layout = UICollectionViewCompositionalLayout.listLayout(groupEstimatedHeight: 50)
    let cvw = UICollectionView(frame: .zero, collectionViewLayout: layout)
    cvw.backgroundColor = .clear
    cvw.translatesAutoresizingMaskIntoConstraints = false
    cvw.register(ShopItemCell.self, forCellWithReuseIdentifier: ShopItemCell.reuseIdentifier)
    cvw.delegate = self
    return cvw
  }()

  private var viewModel: ShopViewModel

  // swiftlint:disable:next implicitly_unwrapped_optional
  private var articlesDataSource: ArticlesDataSource!

  init(withViewModel: ShopViewModel) {
    viewModel = withViewModel
    super.init(nibName: nil, bundle: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("Not happening")
  }

  override func viewDidLoad() {
    super.viewDidLoad()
    title = NSLocalizedString("Purchase", comment: "")
    view.backgroundColor = .systemBackground

    navigationController?.navigationBar.tintColor = .label

    navigationItem.rightBarButtonItem = UIBarButtonItem(
      title: NSLocalizedString("Done", comment: ""),
      style: .plain,
      target: self,
      action: #selector(doneTapped))

    setupViews()

    setupDataSource()
    applyDataSource()
  }

  private func setupViews() {
    view.addSubview(articlesCollectionView)

    NSLayoutConstraint.activate([
      articlesCollectionView.topAnchor.constraint(equalTo: view.safeAreaLayoutGuide.topAnchor, constant: 16),
      articlesCollectionView.leadingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leadingAnchor, constant: 8),
      articlesCollectionView.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -8),
      articlesCollectionView.bottomAnchor.constraint(equalTo: view.bottomAnchor)
    ])
  }

  private func setupDataSource() {
    articlesDataSource = ArticlesDataSource(
      collectionView: articlesCollectionView
    ) { collectionView, indexPath, item -> UICollectionViewCell? in
      let cell = collectionView.dequeueReusableCell(
        withReuseIdentifier: ShopItemCell.reuseIdentifier,
        // swiftlint:disable:next force_cast
        for: indexPath) as! ShopItemCell
      if let articleItem = item as? ShoppingListItem {
        cell.config(with: articleItem)
      }
      return cell
    }
  }

  private func applyDataSource() {
    var snapshot = ArticlesDataSnapshot()

    let listItems = viewModel.currentList.itemsArray.filter {
      $0.article != nil && !$0.didDiscardItem
    }

    var checkedItems = listItems.filter { $0.isPurchased }
    var uncheckedItems = listItems.filter { !$0.isPurchased }

    // swiftlint:disable force_unwrapping
    checkedItems.sort { $0.article!.name! < $1.article!.name! }
    uncheckedItems.sort { $0.article!.name! < $1.article!.name! }
    // swiftlint:enable force_unwrapping

    var finalList = uncheckedItems
    finalList.append(contentsOf: checkedItems)

    let section = Section(
      title: NSLocalizedString("Articles", comment: ""),
      items: finalList)

    snapshot.appendSections([section])
    snapshot.appendItems(section.items, toSection: section)

    articlesDataSource.apply(snapshot)
  }

  @objc private func doneTapped() {
    viewModel.finishPurcahse()
    navigationController?.popViewController(animated: true)
  }
}

extension ShopViewController: UICollectionViewDelegate {
  func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
    // swiftlint:disable:next force_cast
    let itemSelected = articlesDataSource.itemIdentifier(for: indexPath) as! ShoppingListItem

    itemSelected.isPurchased.toggle()

    applyDataSource()
  }
}
