//
//  OnboardingViewController.swift
//  Foodlisting
//
//  Created by Carlos Martinez on 2/4/21.
//

import UIKit

class OnboardingViewController: UIPageViewController {
  private lazy var skipButton: UIButton = {
    let btn = UIButton(type: .custom)
    btn.setTitle(NSLocalizedString("Skip", comment: ""), for: .normal)
    btn.titleLabel?.font = UIFont.newYorkFont(weight: .regular, andTextStyle: .title3)
    btn.titleLabel?.adjustsFontForContentSizeCategory = true
    btn.setTitleColor(.tertiarySystemBackground, for: .normal)
    btn.translatesAutoresizingMaskIntoConstraints = false
    btn.backgroundColor = .clear
    return btn
  }()

  private lazy var nextButton: UIButton = {
    let btn = UIButton(type: .custom)
    btn.setImage(
      UIImage(imageLiteralResourceName: "forwardArrow")
      .withTintColor(.tertiarySystemBackground), for: .normal)
    btn.setTitle(NSLocalizedString("Next ", comment: ""), for: .normal)
    btn.titleLabel?.font = UIFont.newYorkFont(weight: .bold, andTextStyle: .title3)
    btn.titleLabel?.adjustsFontForContentSizeCategory = true
    btn.semanticContentAttribute = .forceRightToLeft
    btn.setTitleColor(.tertiarySystemBackground, for: .normal)
    btn.translatesAutoresizingMaskIntoConstraints = false
    btn.backgroundColor = .clear
    return btn
  }()

  private lazy var startButton: UIButton = {
    let btn = UIButton(type: .custom)
    btn.setTitle(NSLocalizedString("Get started", comment: ""), for: .normal)
    btn.titleLabel?.font = UIFont.newYorkFont(weight: .bold, andTextStyle: .title3)
    btn.titleLabel?.adjustsFontForContentSizeCategory = true
    btn.setTitleColor(.label, for: .normal)
    btn.translatesAutoresizingMaskIntoConstraints = false
    btn.backgroundColor = .tertiarySystemBackground
    btn.isHidden = true
    return btn
  }()

  private let pageControl: UIPageControl = {
    let pageContrl = UIPageControl()
    pageContrl.translatesAutoresizingMaskIntoConstraints = false
    pageContrl.currentPageIndicatorTintColor = .tertiarySystemBackground
    pageContrl.pageIndicatorTintColor = .systemGray2
    return pageContrl
  }()

  private var orderedViewControllers: [UIViewController] = []

  private let viewModel = OnboardingViewModel()

  private var currentIndex = 0 {
    didSet {
      pageControl.currentPage = currentIndex
    }
  }

  init() {
    super.init(transitionStyle: .scroll, navigationOrientation: .horizontal, options: nil)
  }

  required init?(coder: NSCoder) {
    fatalError("init(coder:) has not been implemented")
  }

  override func viewDidLoad() {
    super.viewDidLoad()
    dataSource = self
    delegate = self

    setupControlViews()

    let firstViewController = createViewController(forIndex: 0)
    orderedViewControllers.append(firstViewController)
    setViewControllers([firstViewController], direction: .forward, animated: true)
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)
    navigationController?.setNavigationBarHidden(true, animated: animated)
  }

  override func viewWillDisappear(_ animated: Bool) {
    super.viewWillDisappear(animated)
    navigationController?.setNavigationBarHidden(false, animated: animated)
  }

  private func setupControlViews() {
    view.addSubview(nextButton)
    view.addSubview(skipButton)
    view.addSubview(startButton)
    view.addSubview(pageControl)

    NSLayoutConstraint.activate([
      nextButton.bottomAnchor.constraint(equalTo: view.safeAreaLayoutGuide.bottomAnchor, constant: -16),
      nextButton.trailingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.trailingAnchor, constant: -16),

      skipButton.centerYAnchor.constraint(equalTo: nextButton.centerYAnchor),
      skipButton.leadingAnchor.constraint(equalTo: view.safeAreaLayoutGuide.leadingAnchor, constant: 16),

      startButton.bottomAnchor.constraint(equalTo: nextButton.bottomAnchor),
      startButton.centerXAnchor.constraint(equalTo: view.safeAreaLayoutGuide.centerXAnchor),
      startButton.widthAnchor.constraint(equalTo: view.safeAreaLayoutGuide.widthAnchor, multiplier: 0.8),
      startButton.heightAnchor.constraint(equalToConstant: 50),

      pageControl.bottomAnchor.constraint(equalTo: startButton.topAnchor, constant: -50),
      pageControl.centerXAnchor.constraint(equalTo: view.centerXAnchor),
      pageControl.widthAnchor.constraint(equalTo: view.widthAnchor, multiplier: 0.5)
    ])

    startButton.layer.cornerRadius = 25
    nextButton.addTarget(self, action: #selector(nextAction(sender:)), for: .touchUpInside)
    skipButton.addTarget(self, action: #selector(skipAction(sender:)), for: .touchUpInside)
    startButton.addTarget(self, action: #selector(skipAction(sender:)), for: .touchUpInside)

    pageControl.numberOfPages = viewModel.items.count
    pageControl.currentPage = 0
  }

  private func createViewController(forIndex i: Int) -> OnboardingContentViewController {
    let onboardingItem = viewModel.items[i]
    let newController = OnboardingContentViewController()
    newController.contentItem = onboardingItem
    return newController
  }

  // MARK: Button action handlers
  @objc private func nextAction(sender: UITapGestureRecognizer) {
    guard currentIndex < viewModel.items.count else { return }

    let viewController: UIViewController

    if currentIndex + 1 < orderedViewControllers.count {
      viewController = orderedViewControllers[currentIndex + 1]
    } else {
      viewController = createViewController(forIndex: currentIndex + 1)
      orderedViewControllers.append(viewController)
    }

    currentIndex += 1

    setViewControllers([viewController], direction: .forward, animated: true) { [weak self] _ in
      guard let self = self else { return }
      if self.currentIndex == self.viewModel.items.count - 1 {
        self.skipButton.isHidden = true
        self.nextButton.isHidden = true
        self.startButton.isHidden = false
      }
    }
  }

  @objc private func skipAction(sender: UITapGestureRecognizer) {
    let landingViewController = LandingViewController()
    navigationController?.pushViewController(landingViewController, animated: true)
  }
}


// MARK: - Page Controller Data Source and Delegate
extension OnboardingViewController: UIPageViewControllerDelegate, UIPageViewControllerDataSource {
  func pageViewController(
    _ pageViewController: UIPageViewController,
    viewControllerBefore viewController: UIViewController
  ) -> UIViewController? {
    guard let viewControllerIndex = orderedViewControllers.firstIndex(of: viewController) else { return nil }

    guard viewControllerIndex > 0 else { return nil }

    guard !orderedViewControllers.isEmpty else { return nil }

    return orderedViewControllers[viewControllerIndex - 1]
  }

  func pageViewController(_ pageViewController: UIPageViewController, viewControllerAfter viewController: UIViewController) -> UIViewController? {
    guard let viewControllerIndex = orderedViewControllers.firstIndex(of: viewController) else { return nil }
    guard viewControllerIndex + 1 < viewModel.items.count else { return nil }

    let nextIndex = viewControllerIndex + 1

    if nextIndex < orderedViewControllers.count {
      return orderedViewControllers[nextIndex]
    }

    let newController = createViewController(forIndex: viewControllerIndex + 1)
    orderedViewControllers.append(newController)
    return newController
  }

  func pageViewController(_ pageViewController: UIPageViewController, willTransitionTo pendingViewControllers: [UIViewController]) {
    guard let pendingViewController = pendingViewControllers.first else { return }
    guard let viewControllerIndex = orderedViewControllers.firstIndex(of: pendingViewController) else { return }
    currentIndex = viewControllerIndex
  }

  func pageViewController(
    _ pageViewController: UIPageViewController,
    didFinishAnimating finished: Bool,
    previousViewControllers: [UIViewController],
    transitionCompleted completed: Bool
  ) {
    guard completed else { return }
    guard let previousController = previousViewControllers.first else { return }
    guard let viewControllerIndex = orderedViewControllers.firstIndex(of: previousController) else { return }

    let lastItemIndex = viewModel.items.count - 1
    let isLast = currentIndex == lastItemIndex && viewControllerIndex == lastItemIndex - 1

    skipButton.isHidden = isLast
    nextButton.isHidden = isLast
    startButton.isHidden = !isLast
  }
}
